import asyncio
import importlib

from tortoise import Model, fields, Tortoise
from jija.utils.collector import collect_subclasses
from aerich.ddl.postgres import PostgresDDL


class ModelUpdater:
    connection = None
    ddl = None

    def __init__(self, model):
        self.model = model

    async def update(self, existing_models):
        if self.model._meta.db_table in existing_models:
            await self.update_fields()
        else:
            print('jija')
            await self.create_model()

    async def update_fields(self):
        model_fields = self.model._meta.fields_map
        db_fields = await self.get_db_fields()

        new_fields = list(filter(lambda field_name: not db_fields.get(field_name), model_fields))
        old_fields = list(filter(lambda field_name: not model_fields.get(field_name), db_fields))
        same_fields = list(filter(lambda field_name: model_fields.get(field_name), db_fields))

        for key in same_fields:
            different = FieldGenerator.get_different(db_fields[key], model_fields[key])
            if different:
                print(key, different)

        for key in new_fields:
            await self.__create_field(model_fields[key])

        for key in old_fields:
            await self.__delete_field(db_fields[key])
        # print(create_fields)
        # for name in model_fields.values)
        # print(model_fields)
        # print(db_fields)

    async def __create_field(self, field):
        ddl = self.ddl.add_column(self.model, field)
        await self.connection.execute_query_dict(ddl)

    async def __delete_field(self, field):
        pass

    async def get_db_fields(self):
        raw_db_fields = await self.connection.execute_query_dict(f"""
            SELECT table_name, column_name, column_default, is_nullable, data_type, character_maximum_length
            FROM information_schema.columns
            WHERE table_schema = 'public' and table_name = '{self.model._meta.db_table}'
        """)

        db_fields = {}
        for field_data in raw_db_fields:
            db_fields[field_data['column_name']] = FieldGenerator.generate_field(field_data)
        return db_fields

    async def create_model(self):
        await self.connection.execute_query_dict(self.ddl.create_table(self.model))


class FieldGenerator:
    @classmethod
    def generate_field(cls, data):
        field_type = cls.__get_type(data['data_type'])
        kwargs = cls.__get_kwargs(data)

        return field_type(**kwargs)

    @classmethod
    def __get_type(cls, data_type):
        if data_type == 'integer':
            return fields.IntField
        elif data_type == 'character varying':
            return fields.TextField
        elif data_type == 'text':
            return fields.TextField
        elif data_type == 'double precision':
            return fields.FloatField
        elif data_type == 'timestamp with time zone':
            return fields.DatetimeField
        elif data_type == 'date':
            return fields.DateField
        else:
            raise ValueError('Unknown field type')

    @classmethod
    def __get_kwargs(cls, data):
        kwargs = {
            'null': data['is_nullable'] == 'YES',
        }

        try:
            default = data['column_default']
            if default != f"nextval('{data['table_name']}_id_seq'::regclass)":
                kwargs['default'] = default
        except ValueError:
            pass

        max_length = data.get('character_maximum_length')
        if max_length:
            kwargs['max_length'] = max_length

        return kwargs

    @classmethod
    def get_different(cls, old, new):
        different = []
        for item in ('default', 'null', 'max_length'):
            if hasattr(old, item):
                if str(getattr(old, item)) != str(getattr(new, item)):
                    different.append(item)

        return different

    @classmethod
    def field_to_dict(cls):
        list(diff(old_data_field, new_data_field))


class Jija(Model):
    a = fields.TextField()
    g = fields.IntField()


class Juja(Model):
    a = fields.CharField(max_length=334)
    b = fields.TextField()
    c = fields.IntField(default=5)
    d = fields.FloatField(null=True)
    e = fields.DatetimeField(unique=True)
    f = fields.DateField()

async def main():
    await initdb()
    await Migrator.run()


async def initdb():
    settings = {
        'connections': {
            'default': {
                'engine': 'tortoise.backends.asyncpg',
                'credentials': {
                    'port': '5432',
                    'user': 'postgres',
                    # 'database': 'tg_sniffer',
                    'database': 'test',
                    'host': 'localhost',
                    'password': '4296'
                }
            },
        },

        'apps': {
            'my_app': {
                'models': ['__main__'],
                'default_connection': 'default',
            }
        },

        'use_tz': False,
        'timezone': 'UTC'
    }

    await Tortoise.init(config=settings)

class FakeClient:
    class capabilities:
        inline_comment = None


class Migrator:
    @classmethod
    async def run(cls):
        ModelUpdater.connection = Tortoise.get_connection("default")
        ModelUpdater.ddl = PostgresDDL(FakeClient)

        models = cls.get_models()
        exist_models = await cls.get_existing_models()

        for model in models:
            await model.update(exist_models)
        #     if model._meta.db_table in exist_models:
        #         await cls.migrate_model(model)
        #     else:
        #         await cls.create_model(model)

    @classmethod
    def get_models(cls):
        mm = importlib.import_module('__main__')

        models = []
        for model in collect_subclasses(mm, Model):
            model = ModelUpdater(model)
            models.append(model)

        return models

    @classmethod
    async def get_existing_models(cls):
        conn = Tortoise.get_connection("default")
        res = list(map(
            lambda item: item['table_name'],
            await conn.execute_query_dict(
                "SELECT TABLE_NAME FROM information_schema.tables WHERE table_schema = 'public'")
        ))
        return res

import aerich

if __name__ == '__main__':
    # print(Juja._meta.fields_map)
    asyncio.get_event_loop().run_until_complete(main())



from aiohttp import web


@web.middleware
class Middleware:
    async def handler(self, request, handler):
        raise NotImplementedError()

    async def __call__(self, request, handler):
        return await self.handler(request, handler)



import os.path


class Path:
    def __init__(self, path):
        self.__path, self.__py_ending = self.split(path)

    @staticmethod
    def split(path):
        if isinstance(path, list):
            path = path.copy()

        elif path.count('/'):
            path = path.split('/')

        elif path.count('\\'):
            path = path.split('\\')

        elif path.count('.'):
            path = path.split('.')
            if path[-1] == 'py':
                path[-2] += f'.{path[-1]}'
                del path[-1]

        else:
            path = [path]

        if not path:
            return path, False

        py_ending = path[-1].endswith('.py')
        if py_ending:
            path[-1] = path[-1][:-3]
        return path, py_ending

    @property
    def python(self):
        return '.'.join(self.__path)

    @property
    def system(self):
        return os.path.join(*self.__path) + ('.py' if self.__py_ending else '')

    def has_protected_nodes(self):
        for node in self.__path:
            if node.startswith('__'):
                return True

    def __add__(self, other):
        if isinstance(other, str) and not self.__py_ending:
            return Path(self.__path + [other])

        elif self.__py_ending:
            raise TypeError()

        else:
            raise NotImplementedError()



import datetime
import re

from jija.forms.exceptions import ValidationError


class Validator:
    @classmethod
    async def validate(cls, value, field):
        raise NotImplementedError()


class LengthMinValidator(Validator):
    @classmethod
    async def validate(cls, value, field):
        if field.min_length and len(value) < field.min_length:
            raise ValidationError(f'Строка должна быть больше {field.min_length} символов', value)

        return value


class LengthMaxValidator(Validator):
    @classmethod
    async def validate(cls, value, field):
        if field.max_length and len(value) > field.max_length:
            raise ValidationError(f'Строка должна быть больше {field.max_length} символов', value)

        return value


class RegexValidator(Validator):
    @classmethod
    async def validate(cls, value, field):
        if field.regex and not re.match(field.regex, value):
            raise ValidationError(f'Строка не соответствует шаблону', value)

        return value


class IntegerValidator(Validator):
    @classmethod
    async def validate(cls, value, field):
        try:
            return int(value)
        except ValueError:
            raise ValidationError('Поле должно быть числом', value)


class FloatValidator(Validator):
    @classmethod
    async def validate(cls, value, field):
        try:
            return float(value)
        except ValueError:
            raise ValidationError('Поле должно быть числом', value)


class RangeMinValidator(Validator):
    @classmethod
    async def validate(cls, value, field):
        if field.min_value and value < field.min_value:
            raise ValidationError(f'Число должна быть больше либо равно {field.min_value}', value)

        return value


class RangeMaxValidator(Validator):
    @classmethod
    async def validate(cls, value, field):
        if field.max_value and value > field.max_value:
            raise ValidationError(f'Число должно быть меньше либо равно {field.max_value}', value)

        return value


class DateValidator(Validator):
    @classmethod
    async def validate(cls, value, field):
        try:
            return datetime.datetime.strptime(value, '%Y-%m-%d')
        except ValueError:
            raise ValidationError('Неверный формат', value)


class OptionsValidator(Validator):
    @classmethod
    async def validate(cls, value, field):
        if value not in field.options:
            raise ValidationError(f'Недопустимое значение', value)

        return value



import aerich
from jija.command import Command


class Update(Command):
    async def handle(self):
        from jija.apps import Apps
        from jija.database import DatabaseConfig

        for app_name in Apps.apps:
            command = aerich.Command(tortoise_config=DatabaseConfig.get_config(), app=app_name)
            await command.init()
            await command.upgrade()



import importlib
import os
import sys

import aiohttp_session
from aiohttp import web
from aiohttp_session.cookie_storage import EncryptedCookieStorage

from jija import middlewares
from jija.app import App
from jija import commands
from jija.utils.collector import collect_subclasses
from jija import config
from jija.config.base import Base


class AppGetter(type):
    def __getattr__(self, item):
        app = Apps.apps.get(item)
        if app:
            return app

        raise AttributeError(item)


class Apps(metaclass=AppGetter):
    apps = {}
    commands = {
        'system': commands.COMMANDS
    }

    @classmethod
    def load(cls):
        Apps.apps['core'] = cls.__create_base_app()
        cls.__collect(config.StructureConfig.APPS_PATH, Apps.apps['core'])
        cls.__register_apps()

    @staticmethod
    def __init_configs():
        for config_class in collect_subclasses(config, config.base.Base):
            if not config_class.INITED:
                config_class()

    @classmethod
    def __create_base_app(cls):
        aiohttp_app = web.Application()
        aiohttp_session.setup(aiohttp_app, EncryptedCookieStorage(config.ProjectConfig.SECRET_KEY))

        aiohttp_app.middlewares.extend([
            middlewares.print_request.PrintRequest(),
            middlewares.url_corrector.slash_redirect,
        ])

        if cls.app_exists(config.StructureConfig.CORE_PATH):
            app_class = cls.get_modify_class(config.StructureConfig.CORE_PATH)
        else:
            app_class = App

        app = app_class(path=config.StructureConfig.CORE_PATH, aiohttp_app=aiohttp_app, name='core')

        return app

    @staticmethod
    def app_exists(path):
        return os.path.exists((path + 'app.py').system)

    @classmethod
    def __collect(cls, path, parent):
        if not os.path.exists(path.system):
            return

        for sub_app_name in os.listdir(path.system):

            next_path = path + sub_app_name
            if App.is_app(next_path):
                app = cls.get_modify_class(next_path)(path=next_path, parent=parent, name=sub_app_name)
                cls.commands[sub_app_name] = app.commands
                cls.apps[sub_app_name] = app
                cls.__collect(path + sub_app_name, app)

    @staticmethod
    def get_modify_class(path):
        module = importlib.import_module((path + 'app').python)
        modify_class = list(collect_subclasses(module, App))
        return modify_class[0] if modify_class else App

    @classmethod
    def __register_apps(cls, app=None):
        if not app:
            app = cls.apps['core']

        if not app.childes and app.name != 'core':
            app.parent.aiohttp_app.add_subapp(f'/{app.name}', app.aiohttp_app)

        for child in app.childes:
            cls.__register_apps(child)

    @classmethod
    def get_command(cls, module, command):
        if module is None:
            module = 'system'

        return cls.commands[module][command]

    @classmethod
    def run_command(cls):
        args = sys.argv
        command = args[1].split('.')
        Apps.load()

        if len(command) == 1:
            module = None
            command = command[0]
        else:
            module, command = command

        command_class = cls.get_command(module, command)
        command_obj = command_class()
        command_obj.run()



import datetime
import time

from aiohttp.web_exceptions import HTTPNotFound
from jija.middleware import Middleware


class PrintRequest(Middleware):
    async def handler(self, request, handler):
        timer = time.time()
        try:
            response = await handler(request)
        except HTTPNotFound as exception:
            PrintRequest.print(timer, 404, request.url)
            raise exception

        except Exception as exception:
            PrintRequest.print(timer, 500, request.url)
            raise exception

        PrintRequest.print(timer, response.status, request.url)
        return response

    @classmethod
    def print(cls, timer, status, url):
        timer = time.time() - timer
        time_label = f'{datetime.datetime.today().replace(microsecond=0)}'
        print(f'[{time_label}] [{round(timer, 4)} s] [{status}] {str(url).split("?")[0]}')



from aiohttp import web


@web.middleware
async def slash_redirect(request, handler):
    if request.path[-1] != '/':
        raise web.HTTPFound(request.path + '/')

    response = await handler(request)
    return response



import asyncio
import datetime
import os
import subprocess

from watchdog.events import RegexMatchingEventHandler
from watchdog.observers import Observer

from jija.command import Command


class MultiThreadEvent(asyncio.Event):
    def set(self) -> None:
        self._loop.call_soon_threadsafe(super().set)


class Run(Command):
    def __init__(self):
        super().__init__()
        self.close_event = MultiThreadEvent()
        self.last_reload = datetime.datetime.now()

        self.run_watcher()

    def run_watcher(self):
        event_handler = RegexMatchingEventHandler([r'.*\.py$'])
        event_handler.on_modified = self.modify_callback

        observer = Observer()
        observer.schedule(event_handler, os.getcwd(), recursive=True)
        observer.start()

    def modify_callback(self, event):
        if self.reload_timeout() and not self.close_event.is_set():
            self.last_reload = datetime.datetime.now()
            self.close_event.set()

    def reload_timeout(self):
        return datetime.datetime.now() - self.last_reload > datetime.timedelta(seconds=5)

    async def prepare(self):
        pass

    async def handle(self):
        while True:
            runner = subprocess.Popen(['venv/scripts/python', 'main.py', 'runprocess'])
            await self.close_event.wait()
            runner.kill()
            self.close_event.clear()
            print()



import importlib
import os

from aiohttp import web

from jija.middleware import Middleware
from jija.utils.collector import collect_subclasses
from jija.command import Command


class App:
    def __init__(self, *, name, path, aiohttp_app=None, parent=None):
        self.__parent = parent
        if parent:
            parent.add_child(self)

        self.__path = path
        self.__name = name
        self.__is_core = aiohttp_app is not None

        self.__routes = None
        self.__database = None
        self.__middlewares = None
        self.__commands = None

        self.__aiohttp_app = None

        self.__childes = []
        self.__load(aiohttp_app)

    @property
    def parent(self):
        return self.__parent

    @property
    def name(self):
        return self.__name

    @property
    def routes(self):
        return self.__routes

    @property
    def database(self):
        return self.__database

    @property
    def middlewares(self):
        return self.__middlewares

    @property
    def aiohttp_app(self):
        return self.__aiohttp_app

    @property
    def childes(self):
        return self.__childes

    @property
    def path(self):
        return self.__path

    @property
    def commands(self):
        return self.__commands

    @property
    def database_config(self):
        database_modules = ['aerich.models'] if self.__is_core else []

        if self.__database:
            database_modules.append((self.__path + 'database').python)

        return database_modules

    def __load(self, aiohttp_app=None):
        self.__routes = self.__get_routes(self.__path)
        self.__database = self.__get_database(self.__path)
        self.__middlewares = self.__get_middlewares(self.__path)
        self.__commands = self.__get_commands(self.__path)

        self.__aiohttp_app = self.get_aiohttp_app(aiohttp_app)

    @staticmethod
    def __get_routes(path):
        try:
            routes_module = importlib.import_module((path + 'routes').python)
            if not hasattr(routes_module, 'routes'):
                return []
            return getattr(routes_module, 'routes')

        except ImportError or AttributeError:
            return []

    @staticmethod
    def __get_database(path):
        try:
            return importlib.import_module((path + 'database').python)

        except ImportError:
            return None

    @staticmethod
    def __get_middlewares(path):
        try:
            raw_middlewares = importlib.import_module((path + 'middlewares').python)
            middlewares = collect_subclasses(raw_middlewares, Middleware)
            return list(map(lambda item: item(), middlewares))

        except ImportError or AttributeError:
            return []

    @staticmethod
    def __get_commands(path):
        commands = {}
        commands_path = path + 'commands'
        if os.path.exists(commands_path.system):
            for file in os.listdir(commands_path.system):
                if file.endswith('.py') and not file.startswith('_'):
                    command_path = commands_path + file.replace('.py', '')
                    command_module = importlib.import_module(command_path.python)

                    command = list(collect_subclasses(command_module, Command))
                    if command:
                        commands[file.replace('.py', '')] = command[0]

        return commands

    @staticmethod
    def is_app(path):
        return os.path.isdir(path.system) and \
               os.path.exists((path + 'app.py').system) and \
               not path.has_protected_nodes()

    def get_aiohttp_app(self, aiohttp_app=None):
        aiohttp_app = aiohttp_app or web.Application()

        aiohttp_app.middlewares.extend(self.__middlewares)
        aiohttp_app.add_routes(self.__routes)

        return aiohttp_app

    def add_child(self, child):
        self.__childes.append(child)



import inspect


def collect_subclasses(module, instance):
    members = []
    for name, obj in inspect.getmembers(module):
        if inspect.isclass(obj) and issubclass(obj, instance) and obj != instance:
            line = inspect.getsourcelines(obj)[1]
            members.append((line, obj))

    return map(lambda item: item[1], sorted(members, key=lambda item: item[0]))



import os

import aerich
from jija.command import Command


class Migrate(Command):
    async def handle(self):
        from jija.apps import Apps
        from jija.database import DatabaseConfig

        for app_name in Apps.apps:
            command = aerich.Command(tortoise_config=DatabaseConfig.get_config(), app=app_name)
            if not os.path.exists(os.path.join(os.getcwd(), 'migrations', app_name)):
                await command.init_db(safe=True)
            else:
                await command.init()
                await command.migrate()
